from typing import Final
from mypy_extensions import u8

from librt.internal import (
    Buffer,
    write_bool, read_bool,
    write_str, read_str,
    write_float, read_float,
    write_int, read_int,
    write_tag, read_tag,
    write_bytes, read_bytes,
)

Tag = u8
TAG_A: Final[Tag] = 33
TAG_B: Final[Tag] = 255
TAG_SPECIAL: Final[Tag] = 239


def test_buffer_basic() -> None:
    b = Buffer(b"foo")
    assert b.getvalue() == b"foo"


def test_buffer_empty() -> None:
    b = Buffer(b"")
    write_int(b, 42)
    b = Buffer(b.getvalue())
    assert read_int(b) == 42


def test_buffer_roundtrip() -> None:
    b = Buffer()
    write_str(b, "foo")
    write_bool(b, True)
    write_str(b, "bar" * 1000)
    write_bool(b, False)
    write_float(b, 0.1)
    write_tag(b, TAG_A)
    write_tag(b, TAG_SPECIAL)
    write_tag(b, TAG_B)
    write_int(b, 1)
    write_int(b, 2)
    write_int(b, 2 ** 85)
    write_int(b, 1234512344)
    write_int(b, 1234512345)
    write_bytes(b, b"foobar")
    write_bytes(b, b"abc" * 1000)

    b = Buffer(b.getvalue())
    assert read_str(b) == "foo"
    assert read_bool(b) is True
    assert read_str(b) == "bar" * 1000
    assert read_bool(b) is False
    assert read_float(b) == 0.1
    assert read_tag(b) == TAG_A
    assert read_tag(b) == TAG_SPECIAL
    assert read_tag(b) == TAG_B
    assert read_int(b) == 1
    assert read_int(b) == 2
    assert read_int(b) == 2 ** 85
    assert read_int(b) == 1234512344
    assert read_int(b) == 1234512345
    assert read_bytes(b) == b"foobar"
    assert read_bytes(b) == b"abc" * 1000


def test_buffer_int_size() -> None:
    for i in (-10, -9, 0, 116, 117):
        b = Buffer()
        write_int(b, i)
        assert len(b.getvalue()) == 1
        b = Buffer(b.getvalue())
        assert read_int(b) == i
    for i in (-12345, -12344, -11, 118, 12344, 12345):
        b = Buffer()
        write_int(b, i)
        assert len(b.getvalue()) <= 9  # sizeof(size_t) + 1
        b = Buffer(b.getvalue())
        assert read_int(b) == i


def test_buffer_str_size() -> None:
    for s in ("", "a", "a" * 127):
        b = Buffer()
        write_str(b, s)
        assert len(b.getvalue()) == len(s) + 1
        b = Buffer(b.getvalue())
        assert read_str(b) == s



"""Build script for mypyc C runtime library and C API unit tests.

The tests are written in C++ and use the Google Test framework.
"""

from __future__ import annotations

import os
import subprocess
import sys
from distutils import ccompiler, sysconfig
from typing import Any

from setuptools import Extension, setup
from setuptools.command.build_ext import build_ext

C_APIS_TO_TEST = [
    "init.c",
    "int_ops.c",
    "float_ops.c",
    "list_ops.c",
    "exc_ops.c",
    "generic_ops.c",
    "pythonsupport.c",
]


class BuildExtGtest(build_ext):
    def get_library_names(self) -> list[str]:
        return ["gtest"]

    def run(self) -> None:
        # Build Google Test, the C++ framework we use for testing C code.
        # The source code for Google Test is copied to this repository.
        gtest_dir = os.path.abspath(
            os.path.join(os.path.dirname(__file__), "..", "external", "googletest")
        )
        os.makedirs(self.build_temp, exist_ok=True)
        subprocess.check_call(
            ["make", "-f", os.path.join(gtest_dir, "make", "Makefile"), f"GTEST_DIR={gtest_dir}"],
            cwd=self.build_temp,
        )
        self.library_dirs = [self.build_temp]
        return build_ext.run(self)


if "--run-capi-tests" in sys.argv:
    sys.argv.pop()

    kwargs: dict[str, Any]
    if sys.platform == "darwin":
        kwargs = {"language": "c++"}
        compile_args = []
    else:
        kwargs = {}
        compile_args = ["--std=c++11"]

    setup(
        name="test_capi",
        version="0.1",
        ext_modules=[
            Extension(
                "test_capi",
                ["test_capi.cc"] + C_APIS_TO_TEST,
                depends=["CPy.h", "mypyc_util.h", "pythonsupport.h"],
                extra_compile_args=["-Wno-unused-function", "-Wno-sign-compare"] + compile_args,
                libraries=["gtest"],
                include_dirs=["../external/googletest", "../external/googletest/include"],
                **kwargs,
            )
        ],
        cmdclass={"build_ext": BuildExtGtest},
    )
else:
    # TODO: we need a way to share our preferred C flags and get_extension() logic with
    # mypyc/build.py without code duplication.
    compiler = ccompiler.new_compiler()
    sysconfig.customize_compiler(compiler)
    cflags: list[str] = []
    if compiler.compiler_type == "unix":
        cflags += ["-O3"]
    elif compiler.compiler_type == "msvc":
        cflags += ["/O2"]

    setup(
        ext_modules=[
            Extension(
                "librt.internal",
                [
                    "librt_internal.c",
                    "init.c",
                    "int_ops.c",
                    "exc_ops.c",
                    "pythonsupport.c",
                    "getargsfast.c",
                ],
                include_dirs=["."],
                extra_compile_args=cflags,
            )
        ]
    )
